#include "flash.h"

//// According to Intel datasheet upper 8 bits of command should
// be all 0s or all 1s

#define FLASH_Read_ID      0x0090
#define FLASH_Read_Status  0x0070
#define FLASH_Clear_Status 0x0050
#define FLASH_Status_Ready 0x0080
#define FLASH_Program      0x0040
#define FLASH_Block_Erase  0x0020
#define FLASH_Suspend      0x00B0
#define FLASH_Resume       0x00D0
#define FLASH_Confirm      0x00D0
#define FLASH_Reset        0x00FF

#define FLASH_Config_Setup 0x0060
#define FLASH_Lock_Block   0x0001
#define FLASH_Unlock_Block 0x00D0
  
#define FLASH_BLOCK_SIZE        0x10000
#define FLASH_BOOT_BLOCK_SIZE    0x2000
#define FLASH_BOOT_BLOCK_COUNT   8

#define FLASH_Intel_code   0x89

#define ROM FlashBaseAddress

extern void diag_printf(const char *fmt, ...);
extern void kickWatchdog(void);

extern volatile unsigned char EpldFlashWriteEnable;

static void flash_query(unsigned short *data)
{
  
  EpldFlashWriteEnable = 1;
  ROM[0] = FLASH_Read_ID;
  data[0] = ROM[0];  // Manufacturer code
  data[1] = ROM[1];  // Device identifier
  ROM[0] = FLASH_Reset;
  EpldFlashWriteEnable = 0;
  diag_printf("Id=%04x,%04x\n",data[0],data[1]);
}

static unsigned FlashId;

int flash_hwr_init(void)
{
  unsigned short data[2];
  typedef int code_fun(unsigned short *);

  flash_query(data);
  
  if (data[0] != FLASH_Intel_code) 
    {
      diag_printf("Not Intel = %x\n", data[0]);
      return FLASH_ERR_HWR;
    }

  if ((data[1] == (unsigned short)0x8897) ||
      (data[1] == (unsigned short)0x88C5))
    {
      // 32 Mbit, OK
      FlashId = (data[0] << 16) | data[1];
    }
  else 
    {
      diag_printf("Unknown device type: %04x\n", data[1]);
      return FLASH_ERR_HWR;
    }
  return FLASH_ERR_OK;
}


// Map a hardware status to a package error
static int flash_hwr_map_error(int err)
{
   if (!(err & 0x007E007E))
     return FLASH_ERR_OK;
   
   diag_printf("Err = %x\n", err);
   if (err & 0x00400040) 
     return FLASH_ERR_ERASE_SUSPEND;
   else if (err & 0x00040004) 
     return FLASH_ERR_PROGRAM_SUSPEND;
   else if (err & 0x00020002)
     return FLASH_ERR_PROTECT;
   else if (err & 0x00100010)
     return FLASH_ERR_PROGRAM;
   else if (err & 0x00200020)
     return FLASH_ERR_ERASE;
   else
     return FLASH_ERR_HWR;  // FIXME
}

int get_status_timeout(unsigned t)
{
   unsigned long stat;
   while(((stat = ROM[0]) & FLASH_Status_Ready) != FLASH_Status_Ready)
     {
	kickWatchdog();
	if (--t == 0) 
	  return FLASH_TIMEOUT;
     }
   return flash_hwr_map_error(stat);   
}


static int lock_sector(volatile unsigned short *block)
{
   if (FlashId != 0x008988C5)
     return FLASH_ERR_OK;
   //diag_printf("sector lock %08x\n",block);
   ROM[0] = FLASH_Clear_Status;
   ROM[0] = FLASH_Config_Setup;
   *block = FLASH_Lock_Block;
   return get_status_timeout(5000000);
}


static int unlock_sector(volatile unsigned short *block)
{
   if (FlashId != 0x008988C5)
     return FLASH_ERR_OK;
   // diag_printf("sector unlock %08x\n",block);
   ROM[0] = FLASH_Clear_Status;
   ROM[0] = FLASH_Config_Setup;
   *block = FLASH_Unlock_Block;
   return get_status_timeout(5000000);
}


int flash_erase_sector(volatile unsigned short *block)
{
   int s;
   diag_printf("sector erase %08x\n",block);
   EpldFlashWriteEnable = 1;
   unlock_sector(block);
   ROM[0] = FLASH_Clear_Status;
   ROM[0] = FLASH_Block_Erase;
   *block = FLASH_Confirm;
   s = get_status_timeout(5000000);
   ROM[0] = FLASH_Reset;
   lock_sector(block);
   EpldFlashWriteEnable = 0;
   return s;
}

int flash_erase_region(volatile unsigned short *addr, int count)
{
  volatile unsigned short *sectorBase = ROM;
  int sectorLen = FLASH_BOOT_BLOCK_SIZE;
  int status;
  
  diag_printf("flash_erase_region(%08x,%d)\n",addr,count);
  
  while (sectorBase < addr + count)
    {
      if (sectorBase >= addr)
        {
          status = flash_erase_sector(sectorBase);
          if (status != FLASH_ERR_OK)
            return flash_hwr_map_error(status);
        }
      if (sectorBase >= ROM + FLASH_BOOT_BLOCK_COUNT * FLASH_BOOT_BLOCK_SIZE/(sizeof *ROM))
        sectorLen = FLASH_BLOCK_SIZE;
      
      sectorBase += sectorLen/(sizeof *ROM);
    }
  return FLASH_ERR_OK;
}

int flash_program_region(volatile unsigned short *addr, unsigned short *data, int count)
{
  int s = FLASH_ERR_OK;
  volatile unsigned short *last = (unsigned short*)0;
  int retryCount = 0;
  
  diag_printf("flash_program_buf(%08x,%08x,%d)\n",addr,data,count);
  diag_printf("ROM = %08x\n",FlashBaseAddress);
  
  EpldFlashWriteEnable = 1;
  ROM[0] = FLASH_Clear_Status;
  
  while (count > 0) 
    {
      // crossed an 8K boundary?
      if ((((unsigned long)addr)^((unsigned long)last)) & ~0x1fff)
        {
          //diag_printf("%x\n",addr);
          if (last)
            lock_sector(last);
          unlock_sector(addr);
          last = addr;
        }

      //diag_printf("%x <= %04x ",addr,*data);
      ROM[0] = FLASH_Program;
      *addr = *data;
      if ((s = get_status_timeout(10000000)))
        break;
      ROM[0] = FLASH_Reset;
      
      //diag_printf("%04x\n",*addr);
      if (*addr != *data) 
        {
          diag_printf("%08x: %04x should be %04x\n",addr,*addr,*data);
          if (retryCount > 3)
            {
              s = FLASH_VERIFY_FAIL;
              break;
            }
          else
            {
              diag_printf("retrying\n");
              ++retryCount;
              continue;
            }
        }
      
      ++addr;
      ++data;
      --count;
      retryCount = 0;      
    }
  
  if (last)
    lock_sector(last);
  
  ROM[0] = FLASH_Reset;
  EpldFlashWriteEnable = 0;
  
  return s;
}
